(*
 * Print the AAL.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2001 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open Fc_ir_util
open Aal_type

let print_asm_fir = ref false

(************************************************************************
 * PRINTING
 ************************************************************************)

(*
 * Operators.
 *)
let print_unop op =
   let s =
      match op with
         UMinusIntOp -> "-"
       | UMinusFloatOp -> "-."
       | UCharOfIntOp -> "(char) "
       | UIntOfFloatOp -> "(int) "
       | UFloatOfIntOp -> "(float) "
   in
      Format.print_string s

let print_binop op =
   let s =
      match op with
         PlusIntOp -> "+"
       | PlusFloatOp -> "+."
       | MinusIntOp -> "-"
       | MinusFloatOp -> "-."
       | MulIntOp -> "*"
       | MulFloatOp -> "*."
       | DivIntOp -> "/"
       | DivFloatOp -> "/."
       | RemIntOp -> "mod"
       | BAndIntOp -> "&"
       | BOrIntOp -> "|"
       | LslIntOp -> "<<"
       | LsrIntOp -> ">>u"
       | AsrIntOp -> ">>"
       | XorIntOp -> "^"
   in
      Format.print_string s

let print_relop op =
   let s =
      match op with
         LTIntOp -> "<"
       | LEIntOp -> "<="
       | EQIntOp -> "="
       | ULTIntOp -> "<u"
       | ULEIntOp -> "<=u"
       | LTFloatOp -> "<."
       | LEFloatOp -> "<=."
       | EQFloatOp -> "=."
   in
      Format.print_string s

(*
 * Print an atom.
 *)
let print_atom = function
   AtomChar c -> Format.print_string ("'" ^ Char.escaped c ^ "'")
 | AtomInt i -> Format.print_int i
 | AtomFloat x -> Format.print_float x
 | AtomNil _ -> Format.print_string "nil"
 | AtomVar v -> print_symbol v

let print_atoms atoms =
   ignore (List.fold_left (fun not_first a ->
                 if not_first then
                    Format.print_string ", ";
                 print_atom a;
                 true) false atoms)

(*
 * Print an address.
 *)
let print_addr a =
   Format.print_string "addr(";
   (match a with
       AddrSafe a ->
          print_atom a
     | AddrSafeOff (a1, a2) ->
          print_atom a1;
          Format.print_string "+";
          print_atom a2
     | AddrUnsafeOff (a1, a2, a3) ->
          print_atom a1;
          Format.print_string "+";
          print_atom a2;
          Format.print_string "<";
          print_atom a3
     | AddrSafeOff2 (a1, a2, a3) ->
          print_atom a1;
          Format.print_string "+";
          print_atom a2;
          Format.print_string "+";
          print_atom a3
     | AddrUnsafeOff2 (a1, a2, a3, a4) ->
          print_atom a1;
          Format.print_string "+";
          print_atom a2;
          Format.print_string "+";
          print_atom a3;
          Format.print_string "<";
          print_atom a4
     | AddrSafeMul (a1, a2, i) ->
          print_atom a1;
          Format.print_string "+";
          print_atom a2;
          Format.print_string "*";
          Format.print_int i
     | AddrSafeMulOff (a1, a2, i, a3) ->
          print_atom a1;
          Format.print_string "+";
          print_atom a2;
          Format.print_string "*";
          Format.print_int i;
          Format.print_string "+";
          print_atom a3
     | AddrUnsafeMulOff (a1, a2, i, a3, a4) ->
          print_atom a1;
          Format.print_string "+";
          print_atom a2;
          Format.print_string "*";
          Format.print_int i;
          Format.print_string "+";
          print_atom a3;
          Format.print_string "<";
          print_atom a4);
   Format.print_string ")"

(*
 * Print an operand.
 *)
let print_operand = function
   OpAtom a -> print_atom a
 | OpAddr a -> print_addr a

(*
 * "let" opener.
 *)
let print_let_open v =
   Format.open_hvbox tabstop;
   Format.print_string "let ";
   print_symbol v;
   Format.print_string " =";
   Format.print_space ()

let rec print_let_close e =
   Format.print_string " in";
   Format.close_box ();
   Format.print_space ();
   print_expr e

(*
 * Print the AAL program.
 *)
and print_expr = function
   LetAtom (v, a, e) ->
      print_let_open v;
      print_atom a;
      print_let_close e

 | LetUnop (v, op, a, e) ->
      print_let_open v;
      print_unop op;
      Format.print_space ();
      print_operand a;
      print_let_close e

 | LetBinop (v, op, a1, a2, e) ->
      print_let_open v;
      print_operand a1;
      Format.print_string " ";
      print_binop op;
      Format.print_space ();
      print_operand a2;
      print_let_close e

 | Jump v ->
      Format.print_string "Jump ";
      print_symbol v

 | RJump (vars, op) ->
      Format.print_string "RJump[";
      ignore (List.fold_left (fun first v ->
                    if not first then
                       Format.print_string " ";
                    print_symbol v;
                    false) true vars);
      Format.print_string "] ";
      print_operand op

 | CJump (op, a1, a2, l1, l2) ->
      Format.open_hvbox 0;
      Format.open_hvbox tabstop;
      Format.print_string "if ";
      print_operand a1;
      Format.print_string " ";
      print_relop op;
      Format.print_string " ";
      print_operand a2;
      Format.print_string " then";
      Format.print_space ();
      Format.print_string "Jump ";
      print_symbol l1;
      Format.close_box ();
      Format.print_space ();
      Format.open_hvbox tabstop;
      Format.print_string "else";
      Format.print_space ();
      Format.print_string "Jump ";
      print_symbol l2;
      Format.close_box ();
      Format.close_box ()

 | Call v ->
      Format.print_string "Call ";
      print_symbol v

 | Reserve (vars, v) ->
      Format.open_hvbox tabstop;
      Format.print_string "Reserve[";
      print_symbol v;
      Format.print_string "]";
      List.iter (fun v ->
            Format.print_space ();
            print_symbol v) vars;
      Format.close_box ()

 | LetAddr (v, addr, e) ->
      print_let_open v;
      Format.print_string "&";
      print_addr addr;
      print_let_close e

 | LetMem (v, addr, e) ->
      print_let_open v;
      Format.print_string "*";
      print_addr addr;
      print_let_close e

 | SetMem (addr, a, e) ->
      Format.open_hvbox tabstop;
      Format.print_string "*";
      print_addr addr;
      Format.print_string " <-";
      Format.print_space ();
      print_operand a;
      Format.print_string ";";
      Format.close_box ();
      Format.print_space ();
      print_expr e

 | LetMemB (v, addr, e) ->
      print_let_open v;
      Format.print_string "*(char *)";
      print_addr addr;
      print_let_close e

 | SetMemB (addr, a, e) ->
      Format.open_hvbox tabstop;
      Format.print_string "*(char *)";
      print_addr addr;
      Format.print_string " <-";
      Format.print_space ();
      print_operand a;
      Format.print_string ";";
      Format.close_box ();
      Format.print_space ();
      print_expr e

 | LetExtCall (v, f, args, e) ->
      print_let_open v;
      print_symbol f;
      Format.print_string "(";
      print_atoms args;
      Format.print_string ")";
      print_let_close e

 | CommentNOP ->
      Format.print_string "..."

 | CommentFIR (exp, e) ->
      if !print_asm_fir then
         begin
            Format.open_hvbox tabstop;
            Format.print_string "/* FIR:";
            Format.print_space ();
            Fc_ir_util.print_expr exp;
            Format.close_box ();
            Format.print_space ();
            Format.print_string "*/";
            Format.print_space ();
            print_expr e
         end

(*
 * Print a block.
 *)
let print_block
    { block_global = gflag;
      block_label = label;
      block_code = exp
    } =
   Format.open_hvbox tabstop;
   if gflag then
      Format.print_string "global(";
   print_symbol label;
   if gflag then
      Format.print_string "):"
   else
      Format.print_string ":";
   Format.print_space ();
   print_expr exp;
   Format.close_box ()

(*
 * Print the program.
 *)
let print_prog
    { aal_data_size = data_size;
      aal_init_sym = init_sym;
      aal_strings = strings;
      aal_blocks = blocks
    } =
   Format.open_vbox 0;
   Format.print_string "data_size=";
   Format.print_int data_size;
   Format.print_space ();
   Format.print_string "init_sym=";
   print_symbol init_sym;
   Format.print_space ();
   List.iter (fun (v, _, s) ->
         Format.print_string "string ";
         print_symbol v;
         Format.print_string (" = \"" ^ String.escaped s ^ "\"");
         Format.print_space ()) strings;
   List.iter (fun block ->
         print_block block;
         Format.print_space ()) blocks;
   Format.close_box ()

(************************************************************************
 * VARS
 ************************************************************************)

(*
 * Add the var.
 *)
let vars_var vset v =
   SymbolSet.add v vset

(*
 * Atom.
 *)
let vars_atom vset = function
   AtomChar _
 | AtomInt _
 | AtomFloat _
 | AtomNil _ ->
      vset
 | AtomVar v ->
      vars_var vset v

let vars_atoms vset atoms =
   List.fold_left vars_atom vset atoms

(*
 * Address.
 *)
let vars_addr vset = function
   AddrSafe a ->
      vars_atom vset a
 | AddrSafeOff (a1, a2) ->
      vars_atom (vars_atom vset a1) a2
 | AddrUnsafeOff (a1, a2, a3) ->
      vars_atom (vars_atom (vars_atom vset a1) a2) a3
 | AddrSafeOff2 (a1, a2, a3) ->
      vars_atom (vars_atom (vars_atom vset a1) a2) a3
 | AddrUnsafeOff2 (a1, a2, a3, a4) ->
      vars_atom (vars_atom (vars_atom (vars_atom vset a1) a2) a3) a4
 | AddrSafeMul (a1, a2, _) ->
      vars_atom (vars_atom vset a1) a2
 | AddrSafeMulOff (a1, a2, _, a3) ->
      vars_atom (vars_atom (vars_atom vset a1) a2) a3
 | AddrUnsafeMulOff (a1, a2, _, a3, a4) ->
      vars_atom (vars_atom (vars_atom (vars_atom vset a1) a2) a3) a4

(*
 * Operand.
 *)
let vars_operand vset = function
   OpAtom a ->
      vars_atom vset a
 | OpAddr a ->
      vars_addr vset a

(*
 * Get all the vars in the exp.
 *)
let rec vars_exp vset = function
   LetAtom (v, a, e) ->
      vars_exp (vars_atom (vars_var vset v) a) e
 | LetUnop (v, _, a, e) ->
      vars_exp (vars_operand (vars_var vset v) a) e
 | LetBinop (v, _, a1, a2, e) ->
      vars_exp (vars_operand (vars_operand (vars_var vset v) a2) a2) e
 | Jump _
 | Call _
 | CommentFIR _
 | CommentNOP ->
      vset
 | RJump (vars, op) ->
      vars_operand (List.fold_left vars_var vset vars) op
 | CJump (_, a1, a2, _, _) ->
      vars_operand (vars_operand vset a1) a2
 | Reserve (vars, v) ->
      vars_var (List.fold_left vars_var vset vars) v
 | LetAddr (v, a, e) ->
      vars_exp (vars_addr (vars_var vset v) a) e
 | LetMem (v, a, e) ->
      vars_exp (vars_addr (vars_var vset v) a) e
 | SetMem (a1, a2, e) ->
      vars_exp (vars_operand (vars_addr vset a1) a2) e
 | LetMemB (v, a, e) ->
      vars_exp (vars_addr (vars_var vset v) a) e
 | SetMemB (a1, a2, e) ->
      vars_exp (vars_operand (vars_addr vset a1) a2) e
 | LetExtCall (v, f, args, e) ->
      vars_var (vars_var (vars_atoms (vars_exp vset e) args) f) v

(*
 * Get all the vars in the program.
 *)
let vars_block vset block =
   vars_exp vset block.block_code

(*
 * Get the vars in all the blocks.
 *)
let vars_blocks blocks =
   List.fold_left vars_block SymbolSet.empty blocks

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
