(*
 * The register allocator performs the following steps:
 *    1. Code generation
 *    2. Register assignment
 *    3. Move coalescing
 *
 * ----------------------------------------------------------------
 *
 * Copyright (C) 1999 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * jyh@cs.caltech.edu
 *)
open Printf

open Debug
open Symbol
open Frame_type
open Codegen_type
open Aal_type
open Dataflow


module RegisterAlloc (Frame : FrameSig) (Codegen : CodegenSig
                                                   with type spset = Frame.spset
                                                   with type block = Frame.block) =
struct
   open Frame

   type prog = Frame.prog

   (*
    * Modules.
    *)
   module Dataflow = Dataflow (Frame)
   module IGraph = Dataflow.IGraph
   module NArray = IGraph.NodeArray
   module NTable = IGraph.Table
   module NSet = IGraph.Set

   open Dataflow

   (*
    * Classification for nodes.
    * A node can be:
    *    Precolored: a machine register
    *    Initial: all non-precolored nodes start here
    *    SimpWL: on the simplify worklist
    *    SpillWL: on the spill worklist
    *    Spilled: spilled
    *    Coalesced: coalesced (set equal to another node)
    *    Colored: assigned to a machine register
    *    Stack: placed on the allocation stack
    *
    * We want to support two functions efficiently:
    *    reclassify node class: reclassify a node
    *    nodes_in_class class: all the nodes with a given class
    *
    * Reclassify takes constant time.
    *)
   type node_class =
      NodePrecolored
    | NodeInitial
    | NodeSimpWL
    | NodeFreezeWL
    | NodeSpillWL
    | NodeSpilled
    | NodeCoalesced
    | NodeColored
    | NodeStack

   (*
    * Moves are also classified.
    *    Coalesced: move has been coalesced
    *    Constrained: the nodes in the move interfere
    *    Frozen: move is not being considered for coalescing
    *    WL: on the move worklist
    *    Active: may be considered for coalescing eventually,
    *       but does not pass the George or Briggs test
    *)
   type move_class =
      MoveCoalesced
    | MoveConstrained
    | MoveFrozen
    | MoveWL
    | MoveActive

   (*
    * For each node we keep
    *    node_node: the classification of the node
    *    node_class: the current classification of the node
    *    node_degree: the number of neighbors
    *    node_alias: a new name for the node
    *    node_color: color that is assigned to a node
    *    node_moves: all the moves the node is in
    *    node_neighbors: all the neighbors
    *)
   type node =
      { node_node : IGraph.node;
        mutable node_class : node_class;
        mutable node_degree : int;
        mutable node_alias : node option;
        mutable node_color : var option;
        mutable node_moves : move list;
        mutable node_neighbors : node list;

        (* Linked list *)
        mutable node_pred : node option ref;
        node_succ : node option ref
      }

   (*
    * For the move
    *    move_dst: dst of the move
    *    move_src: src of the move
    *    move_class: current classification of the move
    *)
   and move =
      { move_dst : node;
        move_src : node;
        mutable move_class : move_class;

        (* Linked list *)
        mutable move_pred : move option ref;
        move_succ : move option ref
      }

   (*
    * All this stuff is defined in Appel,
    * "Modern Compiler Implementation in ML",
    * 1998, Cambridge University Press,
    * pages 242-243.
    *)
   type ra =
      { (* Node worklists *)
        ra_precolored : node option ref;
        ra_initial : node option ref;
        ra_simp_wl : node option ref;
        ra_freeze_wl : node option ref;
        ra_spill_wl : node option ref;
        ra_spilled : node option ref;
        ra_coalesced : node option ref;
        ra_colored : node option ref;
        ra_stack : node option ref;

        (* Move worklists *)
        mv_coalesced : move option ref;
        mv_constrained : move option ref;
        mv_frozen : move option ref;
        mv_wl : move option ref;
        mv_active : move option ref;

        (* Explicit representation of the graph *)
        ra_node_vars : var NArray.t;
        ra_edges : (IGraph.node * IGraph.node, bool) Hashtbl.t
      }

   (************************************************************************
    * NODE OPERATIONS
    ************************************************************************)

   (*
    * Get the list pointer for a node class.
    *)
   let node_worklist ra_nodes = function
      NodePrecolored -> ra_nodes.ra_precolored
    | NodeInitial -> ra_nodes.ra_initial
    | NodeSimpWL -> ra_nodes.ra_simp_wl
    | NodeFreezeWL -> ra_nodes.ra_freeze_wl
    | NodeSpillWL -> ra_nodes.ra_spill_wl
    | NodeSpilled -> ra_nodes.ra_spilled
    | NodeCoalesced -> ra_nodes.ra_coalesced
    | NodeColored -> ra_nodes.ra_colored
    | NodeStack -> ra_nodes.ra_stack

   (*
    * Test for empty.
    *)
   let node_list_empty ra cl =
      let l = node_worklist ra cl in
         match !l with
            Some _ -> false
          | None -> true

   (*
    * Get the head of a list.
    *)
   let node_list_head ra cl =
      let l = node_worklist ra cl in
         match !l with
            Some node -> node
          | None -> raise (Invalid_argument "node_list_head")

   (*
    * Iterate through the node list.
    *)
   let node_iter ra cl f =
      let rec iter = function
         Some node ->
            let next = !(node.node_succ) in
               f node;
               iter next
       | None ->
            ()
      in
         iter (!(node_worklist ra cl))

   (*
    * Get successor node.
    *)
   let node_succ node =
      match !(node.node_succ) with
         Some node -> node
       | None -> raise (Invalid_argument "node_succ")

   (*
    * Create a node in a class.
    *)
   let new_node ra graph node cl =
      let l = node_worklist ra cl in
      let next = !l in
      let succ = ref next in
      let neighbors = IGraph.neighbors graph node in

      (* Postpone computation of neighbors *)
      let node' =
         { node_node = node;
           node_class = cl;
           node_degree = 0;
           node_alias = None;
           node_color = None;
           node_neighbors = [];
           node_moves = [];
           node_pred = l;
           node_succ = succ
         }
      in
      let _ =
         (* Link it into the list *)
         l := Some node';
         match next with
            Some next -> next.node_pred <- succ
          | None -> ()
      in
         node'

   (*
    * Reclassify the node.
    *)
   let node_reclassify ra node cl =
      (* Delete the node from its current list *)
      let pred = node.node_pred in
      let succ = !(node.node_succ) in
      let _ =
         pred := succ;
         match succ with
            Some next -> next.node_pred <- pred
          | None -> ()
      in

      (* Add to the new list *)
      let l = node_worklist ra cl in
      let next = !l in
         l := Some node;
         node.node_class <- cl;
         node.node_pred <- l;
         node.node_succ := next;
         match next with
            Some next -> next.node_pred <- node.node_succ
          | None -> ()

   (*
    * Test for equality.
    *)
   let node_eq node1 node2 =
      node1.node_node = node2.node_node

   (*
    * Get the moves for a node.
    *)
   let node_moves ra node =
      List.filter (fun move ->
            match move.move_class with
               MoveActive
             | MoveWL ->
                  true
             | MoveFrozen
             | MoveCoalesced
             | MoveConstrained ->
                  false) node.node_moves

   (*
    * Get the node's neighbors.
    *)
   let node_neighbors ra node =
      List.filter (fun node ->
            match node.node_class with
               NodeStack
             | NodeCoalesced ->
                  false
             | _ ->
                  true) node.node_neighbors

   (*
    * Say if a node is move related.
    *)
   let node_is_move_related node =
      node.node_moves <> []

   (*
    * Union of two node lists.
    *)
   let node_list_union nodes1 nodes2 =
      let rec mem node1 = function
         node2 :: nodes2 ->
            node_eq node1 node2 || mem node1 nodes2
       | [] ->
            false
      in
      let rec collect nodes1 nodes2 =
         match nodes1 with
            node1 :: nodes1 ->
               if mem node1 nodes2 then
                  collect nodes1 nodes2
               else
                  node1 :: collect nodes1 nodes2
          | [] ->
               nodes2
      in
         collect nodes1 nodes2

   (*
    * Get string representation (var) of a node.
    *)
   let var_of_node ra node =
      NArray.get ra.ra_node_vars node.node_node

   let string_of_node ra node =
      string_of_symbol (var_of_node ra node)

   (************************************************************************
    * MOVE OPERATIONS
    ************************************************************************)

   (*
    * Get the list pointer for a node class.
    *)
   let move_worklist ra_moves = function
      MoveCoalesced -> ra_moves.mv_coalesced
    | MoveConstrained -> ra_moves.mv_constrained
    | MoveFrozen -> ra_moves.mv_frozen
    | MoveWL -> ra_moves.mv_wl
    | MoveActive -> ra_moves.mv_active

   (*
    * Get the head of a list.
    *)
   let move_list_head ra cl =
      let l = move_worklist ra cl in
         match !l with
            Some move -> move
          | None -> raise (Invalid_argument "move_list_head")

   (*
    * Iterate through the node list.
    *)
   let move_iter ra cl f =
      let rec iter = function
         Some move ->
            let next = !(move.move_succ) in
               f move;
               iter next
       | None ->
            ()
      in
         iter (!(move_worklist ra cl))

   (*
    * Create a node in a class.
    *)
   let new_move ra_moves dst src cl =
      let l = move_worklist ra_moves cl in
      let next = !l in
      let succ = ref next in
      let node' =
         { move_dst = dst;
           move_src = src;
           move_class = cl;
           move_pred = l;
           move_succ = succ
         }
      in
      let _ =
         (* Link it into the list *)
         l := Some node';
         match next with
            Some next -> next.move_pred <- succ
          | None -> ()
      in
         node'

   (*
    * Reclassify the node.
    *)
   let move_reclassify ra move cl =
      (* Delete the move from its current list *)
      let pred = move.move_pred in
      let succ = !(move.move_succ) in
      let _ =
         pred := succ;
         match succ with
            Some next -> next.move_pred <- pred
          | None -> ()
      in

      (* Add to the new list *)
      let l = move_worklist ra cl in
      let next = !l in
         l := Some move;
         move.move_class <- cl;
         move.move_pred <- l;
         move.move_succ := next;
         match next with
            Some next -> next.move_pred <- move.move_succ
          | None -> ()

   (************************************************************************
    * PRINTING
    ************************************************************************)

   (*
    * Print all the node lists
    * in the register allocator state.
    *)
   let print_ra ra =
      node_iter ra NodePrecolored (fun node ->
            let v = var_of_node ra node in
               match node.node_color with
                  Some v' ->
                     eprintf "Colored: %s->%s%t" (string_of_symbol v) (string_of_symbol v') eflush
                | None ->
                     raise (Invalid_argument "Register_alloc.print_ra"));
      node_iter ra NodeInitial (fun node ->
            eprintf "Initial: %s%t" (string_of_node ra node) eflush);
      node_iter ra NodeSimpWL (fun node ->
            eprintf "Simp WL: %s%t" (string_of_node ra node) eflush);
      node_iter ra NodeFreezeWL (fun node ->
            eprintf "Freeze WL: %s%t" (string_of_node ra node) eflush);
      node_iter ra NodeSpillWL (fun node ->
            eprintf "Spill WL: %s%t" (string_of_node ra node) eflush);
      node_iter ra NodeSpilled (fun node ->
            eprintf "Spilled: %s%t" (string_of_node ra node) eflush);
      node_iter ra NodeCoalesced (fun node ->
            eprintf "Coalesced: %s%t" (string_of_node ra node) eflush);
      node_iter ra NodeStack (fun node ->
            eprintf "Stack: %s%t" (string_of_node ra node) eflush);
      node_iter ra NodeColored (fun node ->
            let v = var_of_node ra node in
               match node.node_color with
                  Some v' ->
                     eprintf "Colored: %s->%s%t" (string_of_symbol v) (string_of_symbol v') eflush
                | None ->
                     raise (Invalid_argument "Register_alloc.print_ra"))

   (************************************************************************
    * CLASSIFY NODES
    ************************************************************************)

   (*
    * Precolored nodes.
    *
    * Number of colors assigned to registers.
    *)
   let max_colors = List.length Frame.registers

   (*
    * Classify the nodes.  You should create
    * nodes for each of the vertices in the graph.
    * Also, you should add all the edges to the
    * ra.ra_edges table.
    *
    * At this point, we don't know what moves the
    * nodes are involved in, so their class is
    * NodeInitial.
    *)
   let classify_nodes ra graph var_table =
      if !debug_regalloc then
         eprintf "Classify nodes%t" eflush;
      let add_edge node1 node2 =
         Hashtbl.add ra.ra_edges (node1, node2) true
      in
      let narray =
         NArray.init graph (fun node ->
            let v = NTable.find var_table node in
               if List.mem v Frame.registers_all then
                  let node' = new_node ra graph node NodePrecolored in
                     node'.node_color <- Some v;
                     node'
               else
                  new_node ra graph node NodeInitial)
      in
         (* Now add all the edges *)
         raise (Failure "NotImplemented")

   (*
    * At this point, all moves have been added to
    * the nodes, so we can classify them as
    *    NodeSpillWL (if node_degree >= max_colors)
    *    NodeFreezeWL (if node_moves <> [])
    *    NodeSimpWL (otherwise)
    *)
   let reclassify_nodes ra =
      raise (Failure "NotImplemented")

   (*
    * Compute on the moves.
    *)
   let classify_moves ra narray graph moves =
      raise (Failure "NotImplemented")

   (*
    * Classify nodes/moves into worklists.
    *)
   let create igraph =
      let { igraph_graph = graph;
            igraph_node_vars = var_table;
            igraph_var_nodes = node_table;
            igraph_moves = moves
          } = igraph
      in
      let _ =
         if !debug_regalloc then
            eprintf "Register alloc create%t" eflush
      in
      let ra =
         { ra_precolored = ref None;
           ra_initial = ref None;
           ra_simp_wl = ref None;
           ra_freeze_wl = ref None;
           ra_spill_wl = ref None;
           ra_spilled = ref None;
           ra_coalesced = ref None;
           ra_colored = ref None;
           ra_stack = ref None;

           mv_coalesced = ref None;
           mv_constrained = ref None;
           mv_frozen = ref None;
           mv_wl = ref None;
           mv_active = ref None;

           ra_node_vars = NArray.init graph (fun node -> NTable.find var_table node);
           ra_edges = Hashtbl.create 237
         }
      in

      (* Classify all the nodes *)
      let narray = classify_nodes ra graph var_table in

         (* Classify all the moves *)
         classify_moves ra narray graph moves;

         (* Reclassify all the nodes *)
         reclassify_nodes ra;

         ra

   (************************************************************************
    * UTILITIES
    ************************************************************************)

   (*
    * Query the graph for an edge.
    *)
   let query ra node1 node2 =
      try Hashtbl.find ra.ra_edges (node1.node_node, node2.node_node) with
         Not_found ->
            false

   (*
    * Add an edge to the graph.
    *)
   let add_edge ra node1 node2 =
      let node1' = node1.node_node in
      let node2' = node2.node_node in
         if not (query ra node1 node2) && node1' <> node2' then
            begin
               Hashtbl.add ra.ra_edges (node1', node2') true;
               Hashtbl.add ra.ra_edges (node2', node1') true;
               if node1.node_class <> NodePrecolored then
                  begin
                     node1.node_neighbors <- node2 :: node1.node_neighbors;
                     node1.node_degree <- succ node1.node_degree
                  end;
               if node2.node_class <> NodePrecolored then
                  begin
                     node2.node_neighbors <- node1 :: node2.node_neighbors;
                     node2.node_degree <- succ node2.node_degree
                  end
            end

   (*
    * Enable the moves in the nodes.
    *)
   let enable_moves ra nodes =
      List.iter (fun node ->
            List.iter (fun move ->
                  if move.move_class = MoveActive then
                     move_reclassify ra move MoveWL) node.node_moves) nodes

   (*
    * Decrement the degree of a node.
    * This may enable some previously blocked moves.
    *)
   let decrement_degree ra node =
      if node.node_class <> NodePrecolored then
         let degree = node.node_degree in
            node.node_degree <- pred degree;
            if degree = max_colors then
               begin
                  enable_moves ra (node :: node_neighbors ra node);
                  if node_is_move_related node then
                     node_reclassify ra node NodeFreezeWL
                  else
                     node_reclassify ra node NodeSimpWL
               end

   (*
    * Get the alias for a node.
    *)
   let rec node_alias node =
      match node.node_alias with
         Some node ->
            node_alias node
       | None ->
            node

   (*
    * Add a node to a worklist.
    *)
   let add_worklist ra node =
      if node.node_class <> NodePrecolored
         && not (node_is_move_related node)
         && node.node_degree < max_colors
      then
         node_reclassify ra node NodeSimpWL

   (************************************************************************
    * SIMPLIFY
    ************************************************************************)

   (*
    * Simplify.
    *)
   let simplify ra =
      raise (Failure "NotImplemented")

   (************************************************************************
    * COALESCE
    ************************************************************************)

   (*
    * George test.
    *)
   let george_test ra u v =
      raise (Failure "NotImplemented")

   (*
    * Briggs test.
    *)
   let briggs_test ra u v =
      raise (Failure "NotImplemented")

   (*
    * Combine two nodes.
    *)
   let combine ra u v =
      raise (Failure "NotImplemented")

   (*
    * Coalesce.
    *)
   let coalesce ra =
      raise (Failure "NotImplemented")

   (************************************************************************
    * FREEZE
    ************************************************************************)

   (*
    * Freeze the moves for a node.
    *)
   let freeze_moves ra node =
      raise (Failure "NotImplemented")

   (*
    * Freeze.
    *)
   let freeze ra =
      raise (Failure "NotImplemented")

   (************************************************************************
    * SPILL
    ************************************************************************)

   (*
    * Select a node for spilling.
    * Try hard to spill only the vars in
    * vars (a SymbolSet.t), because all other
    * vars are temporaries introduced during Codegen.
    *)
   let spill ra vars =
      raise (Failure "NotImplemented")

   (*
    * Add all the spilled nodes to the spill set.
    *)
   let add_spills spset ra =
      let spset = ref spset in
         node_iter ra NodeSpilled (fun node ->
               let node = node_alias node in
               let v = NArray.get ra.ra_node_vars node.node_node in
                  spset := Frame.spset_spill (!spset) v);
         !spset

   (************************************************************************
    * SELECT
    ************************************************************************)

   (*
    * Assign colors to the nodes in the stack.
    *)
   let rec assign_colors ra =
      raise (Failure "NotImplemented")

   (*
    * Add all the precolored nodes to the spset.
    *)
   let set_colors spset ra =
      let spset = ref spset in
         if !debug_regalloc then
            print_ra ra;

         (* Add all the coalesced nodes *)
         node_iter ra NodeCoalesced (fun node ->
               let v = var_of_node ra node in
               let spset' =
                  match node.node_alias with
                     Some node ->
                        Frame.spset_add (!spset) v (var_of_node ra node)
                   | None ->
                        raise (Invalid_argument "Register_alloc.set_colors")
               in
                  spset := spset');

         (* Add all the colored nodes *)
         node_iter ra NodeColored (fun node ->
               let v = var_of_node ra node in
               let spset' =
                  match node.node_color with
                     Some v' ->
                        Frame.spset_add (!spset) v v'
                   | None ->
                        raise (Invalid_argument "Register_alloc.set_colors")
               in
                  spset := spset');
         !spset

   (*
    * Add all the moves to the spset.
    *)
   let add_moves spset ra =
      let spset = ref spset in
         if !debug_regalloc then
            print_ra ra;

         (* Add all the coalesced nodes *)
         node_iter ra NodeCoalesced (fun node ->
               let v = var_of_node ra node in
               let spset' =
                  match node.node_alias with
                     Some node ->
                        Frame.spset_add (!spset) v (var_of_node ra node)
                   | None ->
                        raise (Invalid_argument "Register_alloc.set_colors")
               in
                  spset := spset');

         !spset

   (************************************************************************
    * MAIN PROGRAM
    ************************************************************************)

   (*
    * Worklists tests.
    *)
   let simp_not_done ra =
      !(ra.ra_simp_wl) <> None

   let move_not_done ra =
      !(ra.mv_wl) <> None

   let freeze_not_done ra =
      !(ra.ra_freeze_wl) <> None

   let spill_not_done ra =
      !(ra.ra_spill_wl) <> None

   let spilled_nodes ra =
      !(ra.ra_spilled) <> None

   (*
    * Main program performs the algorithm on page 244.
    *)
   let rec main spset vars blocks =
      if !debug_regalloc then
         begin
            eprintf "Register_alloc.main%t" eflush;
            Codegen.print_blocks stderr blocks
         end;

      (* Generate interference graph *)
      let igraph = Dataflow.create blocks in

      (* Classify nodes into worklists *)
      let ra = create igraph in

      (* Loop until the first spill *)
      let rec loop1 () =
         if simp_not_done ra then
            begin
               simplify ra;
               loop1 ()
            end
         else if move_not_done ra then
            begin
               coalesce ra;
               loop1 ()
            end
         else if freeze_not_done ra then
            begin
               freeze ra;
               loop1 ()
            end
      in

      (* Loop until all worklists are empty *)
      let rec loop2 () =
         if simp_not_done ra then
            begin
               simplify ra;
               loop2 ()
            end
         else if move_not_done ra then
            begin
               coalesce ra;
               loop2 ()
            end
         else if freeze_not_done ra then
            begin
               freeze ra;
               loop2 ()
            end
         else if spill_not_done ra then
            begin
               spill ra vars;
               loop2 ()
            end
      in
      let _ = loop1 () in
      let spset = add_moves spset ra in
      let _ = loop2 () in

         (* Assign colors *)
         assign_colors ra;

         (* If spilled, try again *)
         if spilled_nodes ra then
            let spset = add_spills spset ra in
            let blocks = Codegen.subst_blocks spset blocks in
               main spset vars blocks
         else
            let spset = set_colors spset ra in
            let blocks = Codegen.subst_blocks spset blocks in
               spset, blocks

   (*
    * Compile a program.
    *)
   let compile
       { aal_data_size = data_size;
         aal_init_sym = init_sym;
         aal_strings = strings;
         aal_blocks = blocks;
         aal_closures = closures
       } =

      (* Collect initial spset *)
      let spset = Frame.initial_spset in
      let spset = Frame.spset_stdargs spset in
      let spset =
         List.fold_left (fun spset (v, v_base, _) ->
               match v_base with
                  Some v_base ->
                     let spset = Frame.spset_clabel spset v_base in
                     let spset = Frame.spset_int spset v 0 in
                        spset
                | None ->
                     Frame.spset_clabel spset v) spset strings
      in
      let spset =
         List.fold_left (fun spset { Aal_type.block_label = label } ->
               Frame.spset_clabel spset label) spset blocks
      in
      let blocks = Codegen.compile_blocks spset blocks in
      let vars = Codegen.vars_blocks blocks in

      (* Build the assembly code *)
      let spset, blocks = main spset vars blocks in
         { asm_data_size = data_size;
           asm_reg_size = Frame.spset_size spset;
           asm_init_sym = init_sym;
           asm_strings = strings;
           asm_blocks = blocks;
           asm_closures = closures
         }
end

(*
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
