(*
 * Produce intermediate code.
 * Also performs type checking.
 *
 * ----------------------------------------------------------------
 *
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * jyh@cs.caltech.edu
 *)

open Symbol
open Fc_ir_type
open Fc_ir_exn
open Fc_ir_exn_type
open Fc_ir_env
open Fc_ir_check
open Fc_ir_standardize

(************************************************************************
 * ENVIRONMENTS
 *
 * Wrap the type and variable environment functions with
 * the normal error handling functions.  You don't have to
 * worry too much about this.
 *
 * A type environment (tenv) gives the definition of type
 * identifiers (typedefs).  The tenv_lookup function takes a type
 * identifier (a symbol) and returns the definition.  The tenv_expand
 * function will expand an outermost type identifier.
 *
 * A variable environment gives the type of each bound variable.
 * You will use the variable environment to keep the types of each
 * bound variable.  The venv_lookup function takes a variable
 * (a symbol) and it returns the ype of that variable.
 *
 * The check_types functions are used to compare types.  The
 * check_types function takes two types, and raises an exception
 * if the two types are not equal.
 ************************************************************************)

(*
 * Default exception.
 *)
let raise_exn pos v =
   SemException (pos, v)

let tenv_lookup = tenv_lookup raise_exn
let tenv_expand = tenv_expand raise_exn
let venv_lookup = venv_lookup raise_exn
let venv_lookup_expand = venv_lookup_expand raise_exn

let check_types = check_types raise_exn
let check_types_list = check_types_list raise_exn
let check_fields = check_fields raise_exn

(*
 * Atom pos labels the file.
 *)
let atom_pos pos =
   string_pos "Fc_ir_ast" (atom_pos pos)

(************************************************************************
 * UTILITIES
 *
 * These are utilities that you may find useful.
 ************************************************************************)

(*
 * Convert an AST binop to an IR binop.
 *)
let binop_of_ast_binop = function
   Fc_ast_type.PlusOp -> PlusOp
 | Fc_ast_type.MinusOp -> MinusOp
 | Fc_ast_type.TimesOp -> MulOp
 | Fc_ast_type.DivideOp -> DivOp
 | Fc_ast_type.ModOp -> RemOp
 | Fc_ast_type.BAndOp -> BAndOp
 | Fc_ast_type.BOrOp -> BOrOp
 | Fc_ast_type.BXorOp -> XorOp
 | Fc_ast_type.LShiftOp -> LslOp
 | Fc_ast_type.RShiftOp -> AsrOp
 | Fc_ast_type.EqOp
 | Fc_ast_type.NotEqOp
 | Fc_ast_type.LeOp
 | Fc_ast_type.LtOp
 | Fc_ast_type.GtOp
 | Fc_ast_type.GeOp ->
      raise (Invalid_argument "Fc_ir_ast.binop_of_ast_binop")

(*
 * Relations are only allowed in IfThenElse,
 * so they are not normal binary operators.
 *)
let relop_of_ast_binop = function
   Fc_ast_type.PlusOp
 | Fc_ast_type.MinusOp
 | Fc_ast_type.TimesOp
 | Fc_ast_type.DivideOp
 | Fc_ast_type.ModOp
 | Fc_ast_type.BAndOp
 | Fc_ast_type.BOrOp
 | Fc_ast_type.BXorOp
 | Fc_ast_type.LShiftOp
 | Fc_ast_type.RShiftOp ->
      raise (Invalid_argument "Fc_ir_ast.relop_of_ast_binop")
 | Fc_ast_type.EqOp -> EqOp
 | Fc_ast_type.NotEqOp -> NEqOp
 | Fc_ast_type.LeOp -> LeOp
 | Fc_ast_type.LtOp -> LtOp
 | Fc_ast_type.GtOp -> GtOp
 | Fc_ast_type.GeOp -> GeOp

(*
 * Get field offset and size.
 *)
let type_of_field tenv pos v ty =
   let ty = tenv_expand tenv pos ty in
      match ty with
         TyAny
       | TyChar
       | TyInt
       | TyFloat
       | TyArray _
       | TyFun _ ->
            raise (SemException (pos, NotAStruct ty))
       | TyStruct fields ->
            (try List.assoc v fields with
                Not_found ->
                   raise (SemException (pos, UnboundLabel v)))
       | TyId v ->
            raise (Invalid_argument "Fc_ir_ast.type_of_field")

(************************************************************************
 * TYPES
 *
 * The make_type function builds an IR type from the AST type.
 * The squash variable is a Boolean value.  If true, array dimensions
 * are omitted (as appropriate in the types of function parameters).
 * Otherwise, array dimensions are preserved (as they should be in
 * a struct type defintion).
 ************************************************************************)

(*
 * Make a IR type from an AST type.
 * If the squash flag is set, all array dimensions
 * are omitted.  It is escpecially important to omit
 * array demensions from function argument types.
 *)
let rec make_type squash tenv = function
   Fc_ast_type.TypeBool  _ ->
      TyInt
 | Fc_ast_type.TypeChar  _ ->
      TyChar
 | Fc_ast_type.TypeInt   _ ->
      TyInt
 | Fc_ast_type.TypeFloat _ ->
      TyFloat
 | Fc_ast_type.TypeId (v, pos) ->
      if squash then
         match tenv_lookup tenv (atom_pos pos) v with
            TyArray (ty, Some _) -> TyArray (ty, None)
          | _ -> TyId v
      else
         TyId v
 | Fc_ast_type.TypeArray (ty, len, pos) ->
      (* Length should be a constant *)
      let len =
         match len with
            Some expr ->
               let i =
                  try Fc_ast_eval.eval_int expr with
                     Fc_ast_eval.EvalException _ ->
                        raise (SemException (atom_pos pos, NotAConstant expr))
               in
                  Some i
          | None ->
               None
      in
      let len =
         if squash then
            None
         else
            len
      in
         TyArray (make_type squash tenv ty, len)
 | Fc_ast_type.TypeStruct (fields, _) ->
      TyStruct (make_type_fields tenv fields)
 | Fc_ast_type.TypeFun (ty_vars, ty_res, _) ->
      TyFun (List.map (make_type true tenv) ty_vars, make_type true tenv ty_res)

and make_type_fields tenv = function
   (v, ty, _) :: fields ->
      (v, make_type false tenv ty) :: make_type_fields tenv fields
 | [] ->
      []

(*
 * Default versions.
 * By default, the make_type function omits array dimensions.
 *)
let make_nosquash_type = make_type false
let make_type = make_type true

(************************************************************************
 * SIMPLE FIR
 *
 * These are some simple AST->IR conversion functions.
 ************************************************************************)

(*
 * Scalar operations.
 *)
let make_scalar op ty tenv venv e cont =
   let v = new_symbol_string "make_scalar" in
   let venv = venv_add venv v ty in
   let ty', exp = cont tenv venv v in
   let exp = LetAtom (v, ty, op e, exp) in
      ty', exp

let make_int = make_scalar (fun i -> AtomInt i) TyInt
let make_char = make_scalar (fun c -> AtomChar c) TyChar
let make_float = make_scalar (fun x -> AtomFloat x) TyFloat

(*
 * String allocation.
 *)
let make_string tenv venv s cont =
   let v = new_symbol_string "make_string" in
   let ty = TyArray (TyChar, None) in
   let venv = venv_add venv v ty in
   let ty, exp = cont tenv venv v in
   let exp = LetString (v, s, exp) in
      ty, exp

(************************************************************************
 * IR PRODUCTION
 *
 * Here is the hard part.  You have to implement each of the
 * conversion functions for building the IR from the AST.  The
 * "cont" function represents the "rest" of the expression.
 * See the lab description for details.
 ************************************************************************)

(*
 * Address-of.
 *)
let rec make_addr_of_expr tenv venv pos e cont =
   let pos = string_pos "make_addr_of_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Unary operations.
 *)
and make_unop_expr tenv venv pos op e cont =
   let pos = string_pos "make_unop_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Pre- and postfix operations.
 *)
and make_uarith_expr tenv venv pos op e cont =
   let pos = string_pos "make_uarith_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Binary operations.
 *)
and make_binop_expr tenv venv pos op e1 e2 cont =
   let pos = string_pos "make_binop_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Boolean relations are short-circuit.
 *)
and make_boolop_expr tenv venv pos op e1 e2 cont =
   let pos = string_pos "make_boolop_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Subscript.
 *)
and make_subscript_expr tenv venv pos e1 e2 cont =
   let pos = string_pos "make_subscript_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Struct projection.
 *)
and make_project_expr tenv venv pos e1 label cont =
   let pos = string_pos "make_project_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Function application.
 *)
and make_apply_expr tenv venv pos f args cont =
   let pos = string_pos "make_apply_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Assignment.
 *)
and make_assign_expr tenv venv pos op e1 e2 cont =
   let pos = string_pos "make_assign_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Conditional.
 *)
and make_if_expr tenv venv pos e1 e2 e3 cont =
   (* Wrap the continuation in a function call *)
   let pos = string_pos "make_if_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * For loop.  Just transform it into a while loop.
 *)
and make_for_expr tenv venv pos e1 e2 e3 e4 cont =
   let e =
      Fc_ast_type.SeqExpr (**)
         ([e1; Fc_ast_type.WhileExpr (**)
              (e2, Fc_ast_type.SeqExpr ([e4; e3], pos), pos)], pos)
   in
      make_expr tenv venv e cont

(*
 * While loop.
 * This becomes a recursive function.
 *)
and make_while_expr tenv venv pos e1 e2 cont =
   (* Put the rest of the program in a function called "break" *)
   let pos = string_pos "make_while_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Sequential execution.
 * Have to wrap the continuation in a function
 * so that environment will be correct.
 *)
and make_seq_expr tenv venv pos el cont =
   let pos = string_pos "make_seq_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Return from the current function.
 *)
and make_return_expr tenv venv pos e cont =
   let pos = string_pos "make_return_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Break from the current loop.
 * Just passes control to the "break" function.
 *)
and make_break_expr tenv venv pos cont =
   let pos = string_pos "make_break_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Variable definitions.
 *)
and make_var_defs_expr tenv venv pos defs cont =
   let pos = string_pos "make_var_defs_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Function declaration just modifies the var environment.
 *)
and make_fun_decl_expr tenv venv pos f ty_vars ty_res cont =
   let pos = string_pos "make_fun_decl_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Function definition.
 *)
and make_fun_def_expr tenv venv pos f vars ty_res body cont =
   (* Function type *)
   let pos = string_pos "make_fun_def_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Type definitions.
 *)
and make_type_defs_expr tenv venv pos types cont =
   let pos = string_pos "make_type_defs_expr" pos in
      raise (SemException (pos, NotImplemented))

(*
 * Produce the intermediate code.
 *)
and make_expr tenv venv e cont =
   match e with
      Fc_ast_type.CharExpr (c, _) ->
         make_char tenv venv c cont
    | Fc_ast_type.IntExpr (i, _) ->
         make_int tenv venv i cont
    | Fc_ast_type.FloatExpr (x, _) ->
         make_float tenv venv x cont
    | Fc_ast_type.StringExpr (s, _) ->
         make_string tenv venv s cont
    | Fc_ast_type.VarExpr (v, _) ->
         cont tenv venv v
    | Fc_ast_type.AddrOfExpr (e, pos) ->
         make_addr_of_expr tenv venv (atom_pos pos) e cont
    | Fc_ast_type.UnOpExpr (op, e, pos) ->
         make_unop_expr tenv venv (atom_pos pos) op e cont
    | Fc_ast_type.UArithExpr (op, e, pos) ->
         make_uarith_expr tenv venv (atom_pos pos) op e cont
    | Fc_ast_type.BoolOpExpr (op, e1, e2, pos) ->
         make_boolop_expr tenv venv (atom_pos pos) op e1 e2 cont
    | Fc_ast_type.BinOpExpr (op, e1, e2, pos) ->
         make_binop_expr tenv venv (atom_pos pos) op e1 e2 cont
    | Fc_ast_type.SubscriptExpr (e1, e2, pos) ->
         make_subscript_expr tenv venv (atom_pos pos) e1 e2 cont
    | Fc_ast_type.ProjectExpr (e, v, pos) ->
         make_project_expr tenv venv (atom_pos pos) e v cont
    | Fc_ast_type.ApplyExpr (f, args, pos) ->
         make_apply_expr tenv venv (atom_pos pos) f args cont
    | Fc_ast_type.AssignExpr (op, e1, e2, pos) ->
         make_assign_expr tenv venv (atom_pos pos) op e1 e2 cont
    | Fc_ast_type.IfExpr (e1, e2, e3, pos) ->
         make_if_expr tenv venv (atom_pos pos) e1 e2 e3 cont
    | Fc_ast_type.ForExpr (e1, e2, e3, e4, pos) ->
         make_for_expr tenv venv pos e1 e2 e3 e4 cont
    | Fc_ast_type.WhileExpr (e1, e2, pos) ->
         make_while_expr tenv venv (atom_pos pos) e1 e2 cont
    | Fc_ast_type.SeqExpr (el, pos) ->
         make_seq_expr tenv venv (atom_pos pos) el cont
    | Fc_ast_type.ReturnExpr (e, pos) ->
         make_return_expr tenv venv (atom_pos pos) e cont
    | Fc_ast_type.BreakExpr pos ->
         make_break_expr tenv venv (atom_pos pos) cont
    | Fc_ast_type.VarDefs (defs, pos) ->
         make_var_defs_expr tenv venv (atom_pos pos) defs cont
    | Fc_ast_type.FunDecl (f, ty_vars, ty_res, pos) ->
         make_fun_decl_expr tenv venv (atom_pos pos) f ty_vars ty_res cont
    | Fc_ast_type.FunDef (f, vars, ty_res, body, pos) ->
         make_fun_def_expr tenv venv (atom_pos pos) f vars ty_res body cont
    | Fc_ast_type.TypeDefs (types, pos) ->
         make_type_defs_expr tenv venv (atom_pos pos) types cont

(*
 * Convert a list of expressions.
 *)
and make_list_expr tenv venv el cont =
   let rec loop tenv venv vars = function
      e :: el ->
         make_expr tenv venv e (fun tenv venv v ->
               loop tenv venv (v :: vars) el)
    | [] ->
         cont tenv venv (List.rev vars)
   in
      loop tenv venv [] el

(*
 * Convert an optional expression.
 *)
and make_expr_opt tenv venv e_opt cont =
   match e_opt with
      None ->
         cont tenv venv None
    | Some e ->
         make_expr tenv venv e (fun tenv venv v ->
               cont tenv venv (Some (AtomVar v)))

(************************************************************************
 * GLOBAL FUNCTIONS
 *
 * The make_prog function converts a program (an expr list).
 * The make_expr function converts a single expression.
 * Both functions return the innermost type and variable
 * environment.
 ************************************************************************)

(*
 * Apply to a program.
 *)
let make_prog tenv venv el =
   let tenv_ref = ref tenv in
   let venv_ref = ref venv in
   let rec loop tenv venv vars = function
      e :: el ->
         make_expr tenv venv e (fun tenv venv v ->
               loop tenv venv (v :: vars) el)
    | [] ->
         tenv_ref := tenv;
         venv_ref := venv;
         TyInt, Return (AtomInt 0)
   in
   let ty, exp = loop tenv venv [] el in
   let exp = standardize_expr exp in
      !tenv_ref, !venv_ref, ty, exp

(*
 * Real make_expr with default continuation.
 *)
let make_expr tenv venv e =
   let ref_tenv = ref tenv in
   let ref_venv = ref venv in
   let ty, exp =
      make_expr tenv venv e (fun tenv venv v ->
            let pos = Fc_ast_util.pos_of_expr e in
            let pos = string_pos "make_expr_global" (atom_pos pos) in
            let ty = venv_lookup venv pos v in
               ref_tenv := tenv;
               ref_venv := venv;
               ty, Return (AtomVar v))
   in
   let exp = standardize_expr exp in
      !ref_tenv, !ref_venv, ty, exp

(*
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
