(*
 * Fc_semant exception handling.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Debug
open Symbol
open Fc_ir_util
open Fc_ir_exn_type

(************************************************************************
 * LOCATIONS
 ************************************************************************)

(*
 * Print a location.
 *)
let rec print_exn_loc print_loc = function
   ExnAtom a ->
      print_loc a
 | ExnString (s, a) ->
      print_exn_loc print_loc a;
      Format.print_space ();
      Format.print_string "/";
      Format.print_string s
 | ExnInt (i, a) ->
      print_exn_loc print_loc a;
      Format.print_string ".";
      Format.print_int i

let print_exn_loc print_loc loc =
   print_exn_loc print_loc loc;
   match loc with
      ExnString _
    | ExnInt _ ->
         Format.print_string ":";
         Format.print_space ()
    | ExnAtom _ ->
         ()

(*
 * Positions.
 *)
let debug_pos = ref false
let debug_trace_pos = ref false

let string_pos s pos =
   if !debug_pos then
      begin
         if !debug_trace_pos then
            eprintf "string_pos: %s%t" s eflush;
         ExnString (s, pos)
      end
   else
      pos

let int_pos i pos =
   if !debug_pos then
      ExnInt (i, pos)
   else
      pos

let atom_pos pos =
   ExnAtom pos

(*
 * Semant exception printer.
 *)
let print_sem_exn = function
   UnboundVar v ->
      Format.print_string "unbound variable: ";
      print_symbol v
 | UnboundType v ->
      Format.print_string "unbound type: ";
      print_symbol v
 | UnboundLabel v ->
      Format.print_string "unbound label: ";
      print_symbol v
 | TypeError (ty1, ty2) ->
      Format.open_hvbox tabstop;
      Format.print_string "type error:";
      Format.print_space ();
      Format.print_string "this expression has type";
      Format.print_space ();
      print_type ty2;
      Format.print_space ();
      Format.print_string "but is used with type";
      Format.print_space ();
      print_type ty1;
      Format.close_box ()
 | BinopTypeError (ty1, ty2) ->
      Format.open_hvbox tabstop;
      Format.print_string "type error in binary operation:";
      Format.print_space ();
      Format.print_string "type1:";
      Format.print_space ();
      print_type ty1;
      Format.print_space ();
      Format.print_string "type2:";
      Format.print_space ();
      print_type ty2;
      Format.close_box ()
 | TypeError1 ty1 ->
      Format.open_hvbox tabstop;
      Format.print_string "expected a value of type:";
      Format.print_space ();
      print_type ty1;
      Format.close_box ()
 | NotAPointer ty ->
      Format.print_string "not a pointer:";
      Format.print_space ();
      print_type ty
 | NotAConstant e ->
      Format.print_string "not a constant:";
      Format.print_space ();
      Fc_ast_util.print_expr e
 | NotAScalar ty ->
      Format.print_string "not a scalar:";
      Format.print_space ();
      print_type ty
 | NotAStruct ty ->
      Format.print_string "not a struct:";
      Format.print_space ();
      print_type ty
 | NotAFunction ty ->
      Format.print_string "not a function:";
      Format.print_space ();
      print_type ty
 | NotAnInternalFunction ty ->
      Format.print_string "not an internal function:";
      Format.print_space ();
      print_type ty
 | NotAStructPointer ty ->
      Format.print_string "not a struct pointer: ";
      Format.print_space ();
      print_type ty
 | NotAPrefixValue ty ->
      Format.print_string "invalid argument to prefix operator:";
      Format.print_space ();
      print_type ty
 | NotAddressible e ->
      Format.print_string "invalid argument to '&' operator:";
      Format.print_space ();
      Fc_ast_util.print_expr e
 | ArityMismatch (i, j) ->
      Format.print_string "arity mismatch: wanted ";
      Format.print_int i;
      Format.print_string ", got ";
      Format.print_int j
 | CantInitialize ty ->
      Format.print_string "values of this type can't be initialized: ";
      Format.print_space ();
      print_type ty
 | IRLevel i ->
      Format.print_string "expected IR level ";
      Format.print_int i
 | InternalError s ->
      Format.print_string "internal error: ";
      Format.print_string s
 | InternalTypeError (ty, s) ->
      Format.print_string "internal type error: ";
      Format.print_string s;
      Format.print_space ();
      print_type ty
 | InternalVarError (v, s) ->
      Format.print_string "internal var error: ";
      print_symbol v;
      Format.print_string ": ";
      Format.print_string s
 | SubscriptOutOfBounds i ->
      Format.print_string "subscript out of bounds: ";
      Format.print_int i
 | NotImplemented ->
      Format.print_string "not implemented"
 | PointerDoesNotEscape ->
      Format.print_string "pointer does not escape"

(*
 * Exception printer.
 *)
let print_exn = function
   SemException (loc, exn) ->
      Format.open_hvbox tabstop;
      print_exn_loc Fc_ast_exn.print_pos loc;
      print_sem_exn exn;
      Format.close_box ()
 | IRException (loc, exn) ->
      Format.open_vbox 0;
      print_exn_loc (fun e ->
            Format.open_vbox tabstop;
            Format.print_string "*** Error Location:";
            Format.print_space ();
            print_expr e;
            Format.close_box ();
            Format.print_space ()) loc;
      Format.open_hvbox tabstop;
      Format.print_string "*** Error: ";
      print_sem_exn exn;
      Format.close_box ();
      Format.close_box ()
 | exn ->
      Fc_ast_exn.print_exn exn

let print_exn_chan out exn =
   Format.set_formatter_out_channel out;
   Format.open_vbox 0;
   print_exn exn;
   Format.close_box ();
   Format.print_newline ()

(*
 * Exception handler.
 *)
let catch f x =
   try f x with
      exn ->
         print_exn_chan stderr exn;
         exit 2

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
