(*
 * Type checking.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open Fc_ir_env
open Fc_ir_exn
open Fc_ir_exn_type
open Fc_ir_check

(************************************************************************
 * EXPRESSION CHECKING
 ************************************************************************)

(*
 * Default exception.
 *)
let raise_exn code v =
   IRException (code, v)

let tenv_lookup      = tenv_lookup raise_exn
let tenv_expand      = tenv_expand raise_exn
let venv_lookup      = venv_lookup raise_exn
let venv_lookup_expand = venv_lookup_expand raise_exn

let check_types      = check_types raise_exn
let check_types_list = check_types_list raise_exn
let check_fields     = check_fields raise_exn

(*
 * We use a special symbol for the current return type.
 *)
let return_sym = new_symbol_string "return"
let data_sym   = new_symbol_string "data"
let exit_sym   = new_symbol_string "exit"

let venv_add_return venv ty =
   venv_add venv return_sym ty

let empty_tenv = empty_tenv
let empty_venv = venv_add_return empty_venv TyInt

(*
 * Get the type in the pointer.
 *)
let type_of_deref tenv pos ty =
   match tenv_expand tenv pos ty with
      TyArray (ty, _) -> ty
    | ty -> raise (IRException (pos, NotAPointer ty))

(*
 * Get the type of a field.
 *)
let type_of_field tenv pos ty_struct l =
   let ty_struct = tenv_expand tenv pos ty_struct in
      match ty_struct with
         TyStruct fields ->
            (try
                match List.assoc l fields with
                   TyArray (ty, Some _) -> ty
                 | ty -> ty
             with
                Not_found ->
                   raise (IRException (pos, UnboundLabel l)))
       | _ ->
            raise (IRException (pos, NotAStruct ty_struct))

(*
 * Expand a function type.
 *)
let dest_fun_type tenv pos ty =
   match tenv_expand tenv pos ty with
      TyFun (ty_vars, ty_res) -> ty_vars, ty_res
    | _ -> raise (IRException (pos, NotAFunction ty))

let is_fun_type tenv pos ty =
   match tenv_expand tenv pos ty with
      TyFun (ty_vars, ty_res) -> true
    | _ -> false

(*
 * Get the type of an atom.
 *)
let check_atom venv pos = function
   AtomChar _ -> TyChar
 | AtomInt _ -> TyInt
 | AtomFloat _ -> TyFloat
 | AtomNil ty -> TyArray (ty, None)
 | AtomVar v -> venv_lookup venv pos v

(*
 * General type checking.
 *)
let rec check_expr tenv venv code =
   let pos = string_pos "Fc_ir_typeof" (atom_pos code) in
      match code with
         LetTypes (types, e) ->
            check_types_expr tenv venv pos types e
       | LetFuns (funs, e) ->
            check_funs_expr tenv venv pos funs e
       | LetFuns2 (funs, e) ->
            check_funs2_expr tenv venv pos funs e
       | LetAtom (v, ty, a, e) ->
            check_atom_expr tenv venv pos v ty a e
       | LetCopy (v, ty, a, e) ->
            check_copy_expr tenv venv pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            check_unop_expr tenv venv pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            check_binop_expr tenv venv pos v ty op a1 a2 e
       | LetApply (v, ty, f, args, e) ->
            check_apply_expr tenv venv pos v ty f args e
       | TailCall (f, args) ->
            check_tailcall_expr tenv venv pos f args
       | LetExtCall (v, f, ty, args, e) ->
            check_extcall_expr tenv venv pos v f ty args e
       | IntCall (ty, f, args) ->
            check_intcall_expr tenv venv pos ty f args
       | Return a ->
            check_return tenv venv pos a
       | LetString (v, s, e) ->
            check_string_expr tenv venv pos v s e
       | IfThenElse (op, a1, a2, e1, e2) ->
            check_if_expr tenv venv pos op a1 a2 e1 e2
       | SetVar (v, ty, a, e) ->
            check_set_var_expr tenv venv pos v ty a e
       | LetAddrOfVar (v1, ty, v2, e) ->
            check_addr_of_var_expr tenv venv pos v1 ty v2 e
       | SetSubscript (a1, ty, a2, a3, e) ->
            check_set_subscript_expr tenv venv pos a1 ty a2 a3 e
       | LetSubscript (v, ty, a1, a2, e) ->
            check_subscript_expr tenv venv pos v ty a1 a2 e
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            check_addr_of_subscript_expr tenv venv pos v ty a1 a2 e
       | SetProject (a1, l, ty, a2, e) ->
            check_set_project_expr tenv venv pos a1 l ty a2 e
       | LetProject (v, ty, a, l, e) ->
            check_project_expr tenv venv pos v ty a l e
       | LetAddrOfProject (v, ty, a, l, e) ->
            check_addr_of_project_expr tenv venv pos v ty a l e
       | Memcpy (a1, ty, a2, e) ->
            check_memcpy_expr tenv venv pos a1 ty a2 e
       | LetAlloc (v, ty, e) ->
            check_alloc_expr tenv venv pos v ty e
       | LetClosure (v, ty, f, args, e) ->
            check_closure_expr tenv venv pos v ty f args e

(*
 * Check the type declarations.
 *)
and check_types_expr tenv venv pos types e =
   let pos = string_pos "check_types_expr" pos in
   let tenv = List.fold_left (fun tenv (v, ty) -> tenv_add tenv v ty) tenv types in
      check_expr tenv venv e

(*
 * Check the function declarations.
 *)
and check_funs tenv venv pos funs =
   let pos = string_pos "check_funs" pos in

   (* First, add all the functions to the venv *)
   let venv =
      List.fold_left (fun venv (f, _, f_ty, _, _) ->
            venv_add venv f f_ty) venv funs
   in

   (* Check a single function *)
   let check_fun (f, gflag, f_ty, vars, body) =
      match f_ty with
         TyFun (ty_vars, ty_res) ->
            let len1 = List.length ty_vars in
            let len2 = List.length vars in
            let rec add_vars venv ty_vars vars =
               match ty_vars, vars with
                  ty_v :: ty_vars, v :: vars ->
                     add_vars (venv_add venv v ty_v) ty_vars vars
                | [], [] ->
                     venv
                | _ ->
                     raise (IRException (pos, ArityMismatch (len1, len2)))
            in
            let venv = add_vars venv ty_vars vars in
            let venv =
               match gflag with
                  FunDeclClass
                | FunGlobalClass
                | FunContClass ->
                     venv_add_return venv ty_res
                | FunLocalClass ->
                     venv
            in
            let ty = check_expr tenv venv body in
               check_types tenv pos ty_res ty
       | _ ->
            raise (IRException (pos, NotAFunction f_ty))
   in
      List.iter check_fun funs;
      venv

and check_funs_expr tenv venv pos funs e =
   let pos = string_pos "check_funs_expr" pos in
   let funs =
      List.map (fun (f, gflag, ty, vars, e) ->
            let gflag =
               if gflag then
                  FunGlobalClass
               else
                  FunLocalClass
            in
               f, gflag, ty, vars, e) funs
   in
   let venv = check_funs tenv venv pos funs in
      check_expr tenv venv e

and check_funs2_expr tenv venv pos funs e =
   let pos = string_pos "check_funs_expr" pos in
   let venv = check_funs tenv venv pos funs in
      check_expr tenv venv e

(*
 * Check a LetAtom.
 * Check that the type is the same as the argument.
 *)
and check_atom_expr tenv venv pos v ty a e =
   let pos = string_pos "check_atom_expr" pos in
   let ty' = check_atom venv pos a in
      check_types tenv pos ty ty';
      let venv = venv_add venv v ty in
         check_expr tenv venv e

(*
 * Check a LetCopy.
 * For typechecking, this is similar to LetAtom.
 *)
and check_copy_expr tenv venv pos v ty a e =
   let pos = string_pos "check_copy_expr" pos in
   let _ =
      match a with
         Some a ->
            check_types tenv pos ty (check_atom venv pos a)
       | None ->
            ()
   in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

(*
 * Unary operation.
 *)
and check_unop_expr tenv venv pos v ty op a e =
   let pos = string_pos "check_unop_expr" pos in
   let ty' = tenv_expand tenv pos (check_atom venv pos a) in
   let ty' =
      match op, ty' with
         UMinusOp, (TyChar | TyInt | TyAny) -> TyInt
       | UMinusOp, TyFloat -> TyFloat
       | UNotOp, (TyAny | TyChar | TyInt | TyFloat | TyArray _) -> TyInt
       | UToChar, (TyAny | TyChar | TyInt | TyFloat) -> TyChar
       | UToInt, (TyAny | TyChar | TyInt | TyFloat) -> TyInt
       | UToFloat, (TyAny | TyChar | TyInt | TyFloat) -> TyFloat
       | UToFunPointer, TyFun _ -> TyArray (ty', None)
       | UToFun, TyFun _ -> ty'
       | UToFun, TyArray (ty'', _) ->
            (match tenv_expand tenv pos ty'' with
                TyFun _ -> ty''
              | _ -> raise (IRException (pos, TypeError (ty, ty'))))
       | _ -> raise (IRException (pos, TypeError (ty, ty')))
   in
   let _ = check_types tenv pos ty ty' in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

(*
 * Binary operation.
 *)
and check_binop_expr tenv venv pos v ty op a1 a2 e =
   let pos = string_pos "check_binop_expr" pos in
   let ty1 = tenv_expand tenv (int_pos 1 pos) (check_atom venv pos a1) in
   let ty2 = tenv_expand tenv (int_pos 2 pos) (check_atom venv pos a2) in
   let ty' =
      match op with
         PlusOp ->
            (match ty1, ty2 with
                TyInt, TyInt -> TyInt
              | TyFloat, TyFloat -> TyFloat
              | TyArray (ty, _), TyInt -> TyArray (ty, None)
              | _ -> raise (IRException (pos, BinopTypeError (ty1, ty2))))
       | MinusOp ->
            (match ty1, ty2 with
                TyInt, TyInt -> TyInt
              | TyFloat, TyFloat -> TyFloat
              | TyArray (ty, _), TyInt -> TyArray (ty, None)
              | TyArray (ty1, _), TyArray (ty2, _) ->
                   check_types tenv (int_pos 4 pos) ty1 ty2;
                   TyInt
              | _ -> raise (IRException (pos, BinopTypeError (ty1, ty2))))
       | MulOp
       | DivOp ->
            (match ty1, ty2 with
                TyInt, TyInt -> TyInt
              | TyFloat, TyFloat -> TyFloat
              | _ -> raise (IRException (pos, BinopTypeError (ty1, ty2))))

       | RemOp
       | BAndOp
       | BOrOp
       | LslOp
       | LsrOp
       | AsrOp
       | XorOp ->
            (match ty1, ty2 with
                TyInt, TyInt -> TyInt
              | _ -> raise (IRException (pos, BinopTypeError (ty1, ty2))))
   in
   let _ = check_types tenv (int_pos 3 pos) ty ty' in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

(*
 * Binary relation.
 *)
and check_relop tenv venv pos op a1 a2 =
   let pos = string_pos "check_relop_expr" pos in
   let ty1 = tenv_expand tenv pos (check_atom venv pos a1) in
   let ty2 = tenv_expand tenv pos (check_atom venv pos a2) in
       check_types tenv pos ty1 ty2;
       match ty1 with
          TyAny
        | TyChar
        | TyInt
        | TyFloat
        | TyArray _ ->
             ()
        | TyStruct _
        | TyFun _ ->
             raise (IRException (pos, NotAScalar ty1))
        | TyId _ ->
             raise (Invalid_argument "Fc_ir_typeof.check_relop_expr")

(*
 * Function application.
 * Check that the arguments have the right type and arity.
 *)
and check_tailcall_expr tenv venv pos f args =
   let pos = string_pos "check_tailcall_expr" pos in
   let f_ty = venv_lookup venv pos f in
      check_tailcall_type_expr tenv venv pos f_ty args

and check_tailcall_type_expr tenv venv pos f_ty args =
   let ty_vars, ty_res = dest_fun_type tenv pos f_ty in

   (* Check argument types *)
   let ty_args = List.map (check_atom venv pos) args in
   let len1 = List.length ty_vars in
   let len2 = List.length ty_args in
      if len1 <> len2 then
         raise (IRException (pos, ArityMismatch (len1, len2)));
      List.iter2 (check_types tenv pos) ty_vars ty_args;
      ty_res

and check_apply_expr tenv venv pos v ty f args e =
   let pos = string_pos "check_apply_expr" pos in
   let ty' = check_tailcall_expr tenv venv pos f args in
   let _ = check_types tenv pos ty ty' in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

and check_extcall_expr tenv venv pos v f f_ty args e =
   let pos = string_pos "check_extcall_expr" pos in
   let ty = check_tailcall_type_expr tenv venv pos f_ty args in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

(*
 * Internal call calls the function on the program
 * frame, and the exit continuation.
 *)
and check_intcall_expr tenv venv pos ty f args =
   let pos = string_pos "check_intcall_expr" pos in
   let f_ty = venv_lookup venv pos f in
   let ty_vars, ty_res = dest_fun_type tenv pos f_ty in
      match ty_vars with
         ty_data :: ty_cont :: ty_vars ->
            (match dest_fun_type tenv pos ty_cont with
                [ty_res], TyInt ->
                   let ty_args = List.map (check_atom venv pos) args in
                   let len1 = List.length ty_vars in
                   let len2 = List.length ty_args in
                   let data_type = venv_lookup venv pos data_sym in
                      if len1 <> len2 then
                         raise (IRException (pos, ArityMismatch (len1, len2)));
                      List.iter2 (check_types tenv pos) ty_vars ty_args;
                      check_types tenv pos data_type ty_data;
                      ty_res
              | _ ->
                   raise (IRException (pos, NotAnInternalFunction f_ty)))
       | _ ->
             raise (IRException (pos, NotAnInternalFunction f_ty))

(*
 * Check a return value.
 *)
and check_return tenv venv pos a =
   let pos = string_pos "check_return_expr" pos in
   let ty = check_atom venv pos a in
   let ty' = venv_lookup venv pos return_sym in
      check_types tenv pos ty' ty;
      TyAny

(*
 * Check a string allocation.
 *)
and check_string_expr tenv venv pos v _ e =
   let pos = string_pos "check_string_expr" pos in
   let venv = venv_add venv v (TyArray (TyChar, None)) in
      check_expr tenv venv e

(*
 * Check a conditional.
 *)
and check_if_expr tenv venv pos op a1 a2 e1 e2 =
   let pos = string_pos "check_if_expr" pos in
   check_relop tenv venv pos op a1 a2;
   let ty1 = check_expr tenv venv e1 in
   let ty2 = check_expr tenv venv e2 in
      check_types tenv pos ty1 ty2;
      ty1

(*
 * Set a variable.
 *)
and check_set_var_expr tenv venv pos v ty a e =
   let pos = string_pos "check_set_expr" pos in
   let ty' = check_atom venv pos a in
   let ty_v = venv_lookup venv pos v in
      check_types tenv pos ty ty';
      check_types tenv pos ty_v ty;
      check_expr tenv venv e

(*
 * Address of a variable.
 *)
and check_addr_of_var_expr tenv venv pos v1 ty v2 e =
   let pos = string_pos "check_addr_of_var_expr" pos in
   let ty' = TyArray (venv_lookup venv pos v2, None) in
   let _ = check_types tenv pos ty ty' in
   let venv = venv_add venv v1 ty in
      check_expr tenv venv e

(*
 * Array elements.
 *)
and check_subscript_expr tenv venv pos v ty a1 a2 e =
   let pos = string_pos "check_subscript_expr" pos in
   let ty1 = tenv_expand tenv pos (check_atom venv pos a1) in
   let ty2 = tenv_expand tenv pos (check_atom venv pos a2) in
   let ty' =
      match ty1, ty2 with
         TyArray (ty', _), TyInt -> ty'
       | _ -> raise (IRException (pos, TypeError (ty1, ty2)))
   in
   let _ = check_types tenv pos ty ty' in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

and check_set_subscript_expr tenv venv pos a1 ty a2 a3 e =
   let pos = string_pos "check_set_subscript_expr" pos in
   let ty1 = tenv_expand tenv pos (check_atom venv pos a1) in
   let ty2 = tenv_expand tenv pos (check_atom venv pos a2) in
   let ty3 = check_atom venv pos a3 in
   let ty' =
      match ty1, ty2 with
         TyArray (ty', _), TyInt -> ty'
       | _ -> raise (IRException (pos, TypeError (ty1, ty2)))
   in
      check_types tenv pos ty ty3;
      check_types tenv pos ty ty';
      check_expr tenv venv e

and check_addr_of_subscript_expr tenv venv pos v ty a1 a2 e =
   let pos = string_pos "check_addr_of_subscript_expr" pos in
   let ty1 = tenv_expand tenv pos (check_atom venv pos a1) in
   let ty2 = tenv_expand tenv pos (check_atom venv pos a2) in
   let _ =
      match ty1, ty2 with
         TyArray _, TyInt -> ()
       | _ -> raise (IRException (pos, TypeError (ty1, ty2)))
   in
   let _ = check_types tenv pos ty ty1 in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

(*
 * Struct projection.
 *)
and check_project_expr tenv venv pos v ty a l e =
   let pos = string_pos "check_project_expr" pos in
   let ty_struct = type_of_deref tenv pos (check_atom venv pos a) in
   let ty' = type_of_field tenv pos ty_struct l in
   let _ = check_types tenv pos ty ty' in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

and check_set_project_expr tenv venv pos a1 l ty a2 e =
   let pos = string_pos "check_set_project_expr" pos in
   let ty_struct = type_of_deref tenv pos (check_atom venv pos a1) in
   let ty2 = check_atom venv pos a2 in
   let ty' = type_of_field tenv pos ty_struct l in
      check_types tenv pos ty ty';
      check_types tenv pos ty ty2;
      check_expr tenv venv e

and check_addr_of_project_expr tenv venv pos v ty a l e =
   let pos = string_pos "check_addr_of_project_expr" pos in
   let ty_struct = type_of_deref tenv pos (check_atom venv pos a) in
   let ty' = type_of_field tenv pos ty_struct l in
   let _ = check_types tenv pos ty (TyArray (ty', None)) in
   let venv = venv_add venv v ty in
      check_expr tenv venv e

(*
 * IR2 memory expressions.
 *)
and check_memcpy_expr tenv venv pos a1 ty a2 e =
   let pos = string_pos "check_memcpy_expr" pos in
   let ty1 = check_atom venv pos a1 in
   let ty2 = check_atom venv pos a2 in
      check_types tenv pos ty1 ty2;
      check_types tenv pos ty1 (TyArray (ty, None));
      check_expr tenv venv e

and check_alloc_expr tenv venv pos v ty e =
   let pos = string_pos "check_alloc_expr" pos in
   let venv = venv_add venv v (TyArray (ty, None)) in
      check_expr tenv venv e

and check_closure_expr tenv venv pos v ty f args e =
   let pos = string_pos "check_closure_expr" pos in
   let f_ty = venv_lookup_expand tenv venv pos f in
   let ty_vars, ty_res =
      match f_ty with
         TyFun (ty_vars, ty_res) -> ty_vars, ty_res
       | _ -> raise (IRException (pos, NotAFunction f_ty))
   in

   (* Check argument types *)
   let len1 = List.length ty_vars in
   let len2 = List.length args in
   let _ =
      if len2 > len1 then
         raise (IRException (pos, ArityMismatch (len1, len2)))
   in
   let ty_vars1, ty_vars2 = List_util.split len2 ty_vars in
   let ty_args = List.map (check_atom venv pos) args in
   let pos = int_pos 1 pos in
      List.iter2 (check_types tenv pos) ty_vars1 ty_args;

      (* Check result type *)
      check_types tenv pos ty (TyFun (ty_vars2, ty_res));
      check_expr tenv (venv_add venv v ty) e

(************************************************************************
 * PROGRAM
 ************************************************************************)

(*
 * Add the data segment.
 *)
let check_prog tenv venv
    { prog_data_type = data_ty;
      prog_init_sym = sym_init;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =

   (* Add type definitions *)
   let tenv =
      List.fold_left (fun tenv (v, ty) ->
            tenv_add tenv v ty) tenv types
   in

   (* Add strings *)
   let ty_string = TyArray (TyChar, None) in
   let venv =
      List.fold_left (fun venv (v, _) ->
            venv_add venv v ty_string) venv strings
   in

   (* Add data segment *)
   let venv = venv_add venv data_sym (TyArray (data_ty, None)) in
   let venv = venv_add venv exit_sym (TyFun ([TyAny], TyInt)) in

   (* Check all the funs *)
   let exp = LetFuns2 (funs, TailCall (sym_init, [AtomVar data_sym; AtomVar exit_sym])) in
      check_expr tenv venv exp

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
